/*
 * Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * Copyright (c) 2004-2009 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2008 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

/*
 * Abstract:
 *    Implementation of osm_lid_mgr_t.
 * This file implements the LID Manager object which is responsible for
 * assigning LIDs to all ports on the subnet.
 *
 * DATA STRUCTURES:
 *  p_subn->port_lid_tbl : a vector pointing from lid to its port.
 *  osm db guid2lid domain : a hash from guid to lid (min lid).
 *  p_subn->port_guid_tbl : a map from guid to discovered port obj.
 *
 * ALGORITHM:
 *
 * 0. we define a function to obtain the correct port lid:
 *    lid_mgr_get_port_lid( p_mgr, port, &min_lid ):
 *    0.1 if the port info lid matches the guid2lid return 0
 *    0.2 if the port info has a lid and that range is empty in
 *        port_lid_tbl, return 0 and update the port_lid_tbl and
 *        guid2lid
 *    0.3 else find an empty space in port_lid_tbl, update the
 *    port_lid_tbl and guid2lid, return 1 to flag a change required.
 *
 * 1. During initialization:
 *   1.1 initialize the guid2lid database domain.
 *   1.2 if reassign_lid is not set:
 *   1.2.1 read the persistent data for the domain.
 *   1.2.2 validate no duplicate use of lids and lids are 2^(lmc-1)
 *
 * 2. During SM port lid assignment:
 *   2.1 if reassign_lids is set, make it 2^lmc
 *   2.2 cleanup all port_lid_tbl and re-fill it according to guid2lid
 *   2.3 call lid_mgr_get_port_lid the SM port
 *   2.4 set the port info
 *
 * 3. During all other ports lid assignment:
 *   3.1 go through all ports in the subnet
 *   3.1.1 call lid_mgr_get_port_min_lid
 *   3.1.2 if a change required send the port info
 *   3.2 if any change send the signal PENDING...
 *
 * 4. Store the guid2lid
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif				/* HAVE_CONFIG_H */

#include <stdlib.h>
#include <string.h>
#include <iba/ib_types.h>
#include <complib/cl_qmap.h>
#include <complib/cl_debug.h>
#include <opensm/osm_file_ids.h>
#define FILE_ID OSM_FILE_LID_MGR_C
#include <opensm/osm_lid_mgr.h>
#include <opensm/osm_sm.h>
#include <opensm/osm_log.h>
#include <opensm/osm_node.h>
#include <opensm/osm_switch.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_msgdef.h>
#include <vendor/osm_vendor_api.h>
#include <opensm/osm_db_pack.h>
#include <opensm/osm_inform.h>

/**********************************************************************
  lid range item of qlist
 **********************************************************************/
typedef struct osm_lid_mgr_range {
	cl_list_item_t item;
	uint16_t min_lid;
	uint16_t max_lid;
} osm_lid_mgr_range_t;

void osm_lid_mgr_construct(IN osm_lid_mgr_t * p_mgr)
{
	memset(p_mgr, 0, sizeof(*p_mgr));
}

static void lid_mgr_destroy_free_flid_ranges(IN OUT osm_lid_mgr_t * p_mgr)
{
	cl_list_item_t *p_item;

	while (!cl_is_qlist_empty(&p_mgr->free_flid_ranges)) {
		p_item = cl_qlist_remove_head(&p_mgr->free_flid_ranges);
		free((osm_lid_mgr_range_t *) p_item);
	}
}

void osm_lid_mgr_destroy(IN osm_lid_mgr_t * p_mgr)
{
	cl_list_item_t *p_item;

	OSM_LOG_ENTER(p_mgr->p_log);

	while ((p_item = cl_qlist_remove_head(&p_mgr->free_ranges)) !=
	       cl_qlist_end(&p_mgr->free_ranges))
		free((osm_lid_mgr_range_t *) p_item);

	if (osm_subn_is_flids_enabled(p_mgr->p_subn)) {
		lid_mgr_destroy_free_flid_ranges(p_mgr);
	}

	OSM_LOG_EXIT(p_mgr->p_log);
}

static void lid_mgr_mark_global_flids(IN osm_lid_mgr_t * p_mgr)
{
	OSM_LOG_ENTER(p_mgr->p_log);
	osm_subn_t *p_subn = p_mgr->p_subn;
	/* TODO: we currently work with 16 bits and not 20 */
	/*
	 * mark out the whole global range, so that we do not allow ucast lid
	 * to be assigned with lid from FLID range.
	 */
	uint16_t min_lid = p_subn->opt.global_flid_start;
	uint16_t max_lid = p_subn->opt.global_flid_end;
	uint16_t lid;

	/* mark it as persistent without removing previously marked values */
	for (lid = min_lid; lid <= max_lid; lid++)
		p_mgr->used_lids[lid] |= OSM_RESERVED_LID_PERSISTENT;

	OSM_LOG_EXIT(p_mgr->p_log);
}

/**********************************************************************
Validate the guid to lid data by making sure that under the current
LMC we did not get duplicates. If we do flag them as errors and remove
the entry.
**********************************************************************/
static void lid_mgr_validate_db(IN osm_lid_mgr_t * p_mgr)
{
	cl_qlist_t guids;
	osm_subn_t *p_subn = p_mgr->p_subn;
	osm_db_guid_elem_t *p_item;
	uint64_t sm_port_guid;
	uint16_t lid;
	uint16_t min_lid;
	uint16_t max_lid;
	uint16_t lmc_mask;
	boolean_t lids_ok;
	uint8_t lmc_num_lids = (uint8_t) (1 << osm_subn_get_max_lmc(p_mgr->p_subn));
	uint16_t sm_min_lid;
	uint16_t sm_max_lid;

	OSM_LOG_ENTER(p_mgr->p_log);

	lmc_mask = ~(lmc_num_lids - 1);
	sm_port_guid = cl_ntoh64(p_subn->sm_port_guid);

	cl_qlist_init(&guids);

	if (osm_db_guid2lid_guids(p_mgr->p_g2l, &guids)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0310: "
			"could not get guid list\n");
		goto Exit;
	}

	if (p_subn->opt.master_sm_lid) {
		sm_min_lid = p_subn->opt.master_sm_lid;
		sm_max_lid = sm_min_lid + lmc_num_lids - 1;
	} else {
		sm_min_lid = sm_max_lid = 0;
	}

	while ((p_item = (osm_db_guid_elem_t *) cl_qlist_remove_head(&guids))
	       != (osm_db_guid_elem_t *) cl_qlist_end(&guids)) {
		if (osm_db_guid2lid_get(p_mgr->p_g2l, p_item->guid,
					&min_lid, &max_lid))
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0311: "
				"could not get lid for guid:0x%016" PRIx64 "\n",
				p_item->guid);
		else {
			lids_ok = TRUE;

			if (min_lid > max_lid || min_lid == 0
			    || p_item->guid == 0
			    || max_lid > p_mgr->p_subn->max_ucast_lid_ho) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0312: "
					"Illegal LID range [%u:%u] for "
					"guid:0x%016" PRIx64 "\n", min_lid,
					max_lid, p_item->guid);
				lids_ok = FALSE;
			} else if (min_lid != max_lid
				   && (min_lid & lmc_mask) != min_lid) {
				/* check that if the lids define a range that is
				   valid for the current LMC mask */
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0313: "
					"LID range [%u:%u] for guid:0x%016"
					PRIx64
					" is not aligned according to mask:0x%04x\n",
					min_lid, max_lid, p_item->guid,
					lmc_mask);
				lids_ok = FALSE;
			} else if (p_item->guid != sm_port_guid && sm_min_lid &&
				   ((min_lid <= sm_min_lid && sm_min_lid <= max_lid) ||
				    (sm_min_lid <= min_lid && min_lid <= sm_max_lid))) {
				/* ignore entires that will conflict with master SM LID */
				OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE,
					"LID range [%u:%u] of guid 0x%016" PRIx64
					" conflicts with forced master SM LID %u\n",
					min_lid, max_lid, p_item->guid,
					p_subn->opt.master_sm_lid);
				lids_ok = FALSE;
			} else if (p_subn->opt.master_sm_lid &&
				   p_item->guid == sm_port_guid &&
				   min_lid != p_subn->opt.master_sm_lid) {
				/* ignore entires that will contradict master SM LID */
				OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE,
					"LID range [%u:%u] of guid 0x%016" PRIx64
					" contradicts with forced master SM LID %u\n",
					min_lid, max_lid, p_item->guid,
					p_subn->opt.master_sm_lid);
				lids_ok = FALSE;
			} else {
				/* check if the lids were not previously assigned */
				for (lid = min_lid; lid <= max_lid; lid++) {
					if (p_mgr->used_lids[lid]) {
						OSM_LOG(p_mgr->p_log,
							OSM_LOG_ERROR,
							"ERR 0314: "
							"0x%04x for guid:0x%016"
							PRIx64
							" was previously used\n",
							lid, p_item->guid);
						lids_ok = FALSE;
					}
				}
			}

			if (lids_ok)
				/* mark that it was visited */
				for (lid = min_lid; lid <= max_lid; lid++) {
					if (lid < min_lid+lmc_num_lids)
						p_mgr->used_lids[lid] = OSM_RESERVED_LID_PERSISTENT;
				}
			else if (osm_db_guid2lid_delete(p_mgr->p_g2l,
							p_item->guid))
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0315: failed to delete entry for "
					"guid:0x%016" PRIx64 "\n",
					p_item->guid);
		}		/* got a lid */
		free(p_item);
	}			/* all guids */
Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
}

/*
 * Validate the guid to flid data by making sure that
 * we did not get duplicates. If we do flag them as errors and remove the entry.
 */
static void lid_mgr_validate_flid_db(IN osm_lid_mgr_t * p_mgr)
{
	cl_qlist_t guids;
	osm_db_guid_elem_t *p_item;
	uint16_t flid;
	boolean_t lids_ok;
	uint8_t lmc_num_lids = (uint8_t) (1 << osm_subn_get_max_lmc(p_mgr->p_subn));
	uint16_t sm_min_lid;
	uint16_t sm_max_lid;

	OSM_LOG_ENTER(p_mgr->p_log);

	if (p_mgr->p_subn->opt.master_sm_lid) {
		sm_min_lid = p_mgr->p_subn->opt.master_sm_lid;
		sm_max_lid = sm_min_lid + lmc_num_lids - 1;
	} else {
		sm_min_lid = sm_max_lid = 0;
	}

	cl_qlist_init(&guids);

	if (osm_db_guid2lid_guids(p_mgr->p_g2fl, &guids)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 030D: "
			"could not get guid list\n");
		goto Exit;
	}

	while ((p_item = (osm_db_guid_elem_t *) cl_qlist_remove_head(&guids)) !=
		(osm_db_guid_elem_t *) cl_qlist_end(&guids)) {
		if (osm_db_guid2flid_get(p_mgr->p_g2fl, p_item->guid, &flid)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 030E: "
				"could not get flid for guid:0x%016" PRIx64 "\n",
				p_item->guid);
		} else {
			lids_ok = TRUE;
			/* check that the FLID is inside the defined local range */
			if (flid < p_mgr->p_subn->opt.local_flid_start ||
			    flid > p_mgr->p_subn->opt.local_flid_end) {
				lids_ok = FALSE;
				OSM_LOG(p_mgr->p_log,
					OSM_LOG_ERROR,
					"ERR 0321: "
					"0x%04x for guid:0x%016"
					PRIx64
					" is not in the FLID local range "
					"[0x%" PRIx16 "-0x%" PRIx16"]\n",
					flid, p_item->guid,
					p_mgr->p_subn->opt.local_flid_start,
					p_mgr->p_subn->opt.local_flid_end);
			}
			/* check if the FLID was not previously assigned */
			else if (p_mgr->used_lids[flid]) {
				OSM_LOG(p_mgr->p_log,
					OSM_LOG_ERROR,
					"ERR 030F: "
					"0x%04x for guid:0x%016"
					PRIx64
					" was previously used\n",
					flid, p_item->guid);
				lids_ok = FALSE;
			}

			if (lids_ok) {
				/* the FLID is ok, mark that is used */
				p_mgr->used_lids[flid] = OSM_RESERVED_LID_PERSISTENT;
			} else if (osm_db_guid2flid_delete(p_mgr->p_g2fl, p_item->guid)) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0320: failed to delete entry for "
					"guid:0x%016" PRIx64 "\n",
					p_item->guid);
			}
			/* check that the master_sm_lid doesn't conflict with FLIDs from the db */
			if (sm_min_lid && (sm_min_lid <= flid && flid <= sm_max_lid)) {
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0322: the master_sm_lid %u with lmc num lids %u "
					"conflicts with FLID of guid 0x%" PRIx64 " from guid2flid db. "
					"Set master_sm_lid to 0.\n",
					sm_min_lid, lmc_num_lids, p_item->guid);
				p_mgr->p_subn->opt.master_sm_lid = 0;
			}
		}
		free(p_item);
	}

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
}

static void lid_mgr_reset_flid_free_range(IN osm_lid_mgr_t * p_mgr)
{
	uint16_t lid;
	uint16_t max_lid;
	osm_lid_mgr_range_t *p_range = NULL;

	/* we need to cleanup the empty ranges list */
	lid_mgr_destroy_free_flid_ranges(p_mgr);

	/* The last range is to be up to the max allowed lid. */
	max_lid = p_mgr->p_subn->opt.local_flid_end > p_mgr->p_subn->max_ucast_lid_ho ?
			p_mgr->p_subn->max_ucast_lid_ho :
			p_mgr->p_subn->opt.local_flid_end;

	/*
	 * Construct free FLIDs ranges according to p_mgr->used_lids.
	 */
	for (lid = p_mgr->p_subn->opt.local_flid_start; lid <= max_lid; lid++) {
		if (p_mgr->used_lids[lid] == 0) {
			if (p_range)
				p_range->max_lid = lid;
			else {
				p_range = calloc(1, sizeof(osm_lid_mgr_range_t));
				if (p_range) {
					p_range->min_lid = lid;
					p_range->max_lid = lid;
				}
			}
		} else {
			p_mgr->max_flid_assigned = lid;

			/* this lid is used so we need to finalize the previous free range */
			if (p_range) {
				cl_qlist_insert_tail(&p_mgr->free_flid_ranges,
						     &p_range->item);
				OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
					"new free flid range [%u:%u]\n",
					p_range->min_lid, p_range->max_lid);
				p_range = NULL;
			}
		}
	}


	if (p_range) {
		p_range->max_lid = max_lid;
		cl_qlist_insert_tail(&p_mgr->free_flid_ranges, &p_range->item);
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"final free flid range [%u:%u]\n",
			p_range->min_lid, p_range->max_lid);
	}
}

ib_api_status_t osm_lid_mgr_init(IN osm_lid_mgr_t * p_mgr, IN osm_sm_t * sm)
{
	ib_api_status_t status = IB_SUCCESS;

	OSM_LOG_ENTER(sm->p_log);

	osm_lid_mgr_construct(p_mgr);

	p_mgr->sm = sm;
	p_mgr->p_log = sm->p_log;
	p_mgr->p_subn = sm->p_subn;
	p_mgr->p_db = sm->p_db;
	p_mgr->p_lock = sm->p_lock;

	/* we initialize and restore the db domain of guid to lid map */
	p_mgr->p_g2l = osm_db_domain_init(p_mgr->p_db, "guid2lid");
	if (!p_mgr->p_g2l) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0316: "
			"Error initializing Guid-to-Lid persistent database\n");
		status = IB_ERROR;
		goto Exit;
	}

	cl_qlist_init(&p_mgr->free_ranges);

	if (osm_subn_is_flids_enabled(p_mgr->p_subn)) {
		p_mgr->p_g2fl = osm_db_domain_init(p_mgr->p_db, "guid2flid");
		if (!p_mgr->p_g2fl) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 030A: "
					"Error initializing Guid-to-FLID persistent database\n");
			status = IB_ERROR;
			goto Exit;
		}

		cl_qlist_init(&p_mgr->free_flid_ranges);

		if (osm_db_restore(p_mgr->p_g2fl)) {
#ifndef __WIN__
			/*
			 * When Windows is BSODing, it might corrupt files that
			 * were previously opened for writing, even if the files
			 * are closed, so we might see corrupted guid2lid file.
			 */
			if (p_mgr->p_subn->opt.exit_on_fatal) {
				osm_log_v2(p_mgr->p_log, OSM_LOG_SYS, FILE_ID,
						"FATAL: Error restoring Guid-to-FLID "
						"persistent database\n");
				status = IB_ERROR;
				goto Exit;
			} else
#endif
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
						"ERR 030B: Error restoring Guid-to-FLID "
						"persistent database\n");
		}

		/* validate g2fl */
		lid_mgr_validate_flid_db(p_mgr);

		/* update the free FLID range */
		lid_mgr_reset_flid_free_range(p_mgr);

		/*
		 * Mark global flids in order to avoid using them for unicast lids.
		 * Restoring of the guid 2 flid db, may have marked
		 * only part of the global range.
		 */
		lid_mgr_mark_global_flids(p_mgr);
	}

	/* we use the stored guid to lid table if not forced to reassign */
	if (!p_mgr->p_subn->opt.reassign_lids) {
		if (osm_db_restore(p_mgr->p_g2l)) {
#ifndef __WIN__
			/*
			 * When Windows is BSODing, it might corrupt files that
			 * were previously opened for writing, even if the files
			 * are closed, so we might see corrupted guid2lid file.
			 */
			if (p_mgr->p_subn->opt.exit_on_fatal) {
				osm_log_v2(p_mgr->p_log, OSM_LOG_SYS, FILE_ID,
					   "FATAL: Error restoring Guid-to-Lid "
					   "persistent database\n");
				status = IB_ERROR;
				goto Exit;
			} else
#endif
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0317: Error restoring Guid-to-Lid "
					"persistent database\n");
		}

		/* we need to make sure we did not get duplicates with
		   current lmc */
		lid_mgr_validate_db(p_mgr);
	}


	osm_vlid_mgr_init(&p_mgr->vlid_mgr);

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return status;
}

static uint16_t trim_lid(IN uint16_t lid)
{
	if (lid > IB_LID_UCAST_END_HO || lid < IB_LID_UCAST_START_HO)
		return 0;
	return lid;
}

static void lid_mgr_reset_used_lids(IN osm_lid_mgr_t * p_mgr)
{
	memset(p_mgr->used_lids, 0, sizeof(p_mgr->used_lids));
	/* make sure that the global flid range is set */
	lid_mgr_mark_global_flids(p_mgr);
}

/**********************************************************************
 initialize the manager for a new sweep:
 scans the known persistent assignment and port_lid_tbl
 re-calculate all empty ranges.
 cleanup invalid port_lid_tbl entries
**********************************************************************/
static int lid_mgr_init_sweep(IN osm_lid_mgr_t * p_mgr)
{
	cl_ptr_vector_t *p_discovered_vec = &p_mgr->p_subn->port_lid_tbl;
	uint16_t disc_min_lid, disc_max_lid, db_min_lid, db_max_lid;
	int status = 0;
	cl_list_item_t *p_item;
	boolean_t is_free;
	osm_lid_mgr_range_t *p_range = NULL;
	osm_port_t *p_port;
	osm_vport_t *p_vport;
	cl_qmap_t *p_port_guid_tbl;
	uint8_t lmc_num_lids = (uint8_t) (1 << osm_subn_get_max_lmc(p_mgr->p_subn));
	uint16_t lmc_mask, num_lids, lid;
	uint16_t index, top_index;
	ib_net64_t sm_port_guid = p_mgr->p_subn->sm_port_guid;
	uint16_t sm_min_lid;
	uint16_t sm_max_lid;
	uint16_t global_flid_start = p_mgr->p_subn->opt.global_flid_start;
	uint16_t global_flid_end = p_mgr->p_subn->opt.global_flid_end;

	OSM_LOG_ENTER(p_mgr->p_log);
	lmc_mask = ~(lmc_num_lids - 1);

	if (p_mgr->p_subn->opt.master_sm_lid) {
		sm_min_lid = p_mgr->p_subn->opt.master_sm_lid;
		sm_max_lid = sm_min_lid + lmc_num_lids - 1;
	} else {
		sm_min_lid = sm_max_lid = 0;
	}


	/* We must discard previous guid2lid db if this is the first master
	 * sweep and reassign_lids option is TRUE.
	 * If we came out of standby and honor_guid2lid_file option is TRUE, we
	 * must restore guid2lid db. Otherwise if honor_guid2lid_file option is
	 * FALSE we must discard previous guid2lid db.
	 */
	if (p_mgr->p_subn->first_time_master_sweep == TRUE &&
	    p_mgr->p_subn->opt.reassign_lids == TRUE) {
		osm_db_clear(p_mgr->p_g2l);
		lid_mgr_reset_used_lids(p_mgr);
	} else if (p_mgr->p_subn->coming_out_of_standby == TRUE) {
		osm_db_clear(p_mgr->p_g2l);

		/* clear used lids */
		memset(p_mgr->used_lids, 0, sizeof(p_mgr->used_lids));
		/* FLIDs are enabled */
		if (osm_subn_is_flids_enabled(p_mgr->p_subn)) {
			osm_db_clear(p_mgr->p_g2fl);
			if (osm_db_restore(p_mgr->p_g2fl))
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 030C: "
					"Error restoring Guid-to-FLID "
					"persistent database. Ignoring it\n");
			/* validate g2fl */
			lid_mgr_validate_flid_db(p_mgr);

			lid_mgr_reset_flid_free_range(p_mgr);

			/*
			 * Mark global flids in order to avoid using them for unicast lids.
			 * Restoring of the guid 2 flid db, may have marked
			 * only part of the global range.
			 */
			lid_mgr_mark_global_flids(p_mgr);
		}

		if (p_mgr->p_subn->opt.honor_guid2lid_file == FALSE)
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"Ignore guid2lid file when coming out of standby\n");
		else {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"Honor current guid2lid file when coming out "
				"of standby\n");
			if (osm_db_restore(p_mgr->p_g2l))
				OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR,
					"ERR 0306: "
					"Error restoring Guid-to-Lid "
					"persistent database. Ignoring it\n");
			lid_mgr_validate_db(p_mgr);
		}
	} else if (p_mgr->p_subn->sm_port_changed == TRUE) {
		lid_mgr_reset_used_lids(p_mgr);
		lid_mgr_validate_db(p_mgr);
	} else {
		/* clear "discovered LID" bit form p_mgr->used_lids vector */
		for (lid = 1; lid <= IB_LID_UCAST_END_HO; lid++)
			p_mgr->used_lids[lid] &= OSM_RESERVED_LID_PERSISTENT;
	}

	/* we need to cleanup the empty ranges list */
	while ((p_item = cl_qlist_remove_head(&p_mgr->free_ranges)) !=
	       cl_qlist_end(&p_mgr->free_ranges))
		free((osm_lid_mgr_range_t *) p_item);

	/* first clean up the port_by_lid_tbl */
	for (lid = 0; lid < cl_ptr_vector_get_size(p_discovered_vec); lid++)
		cl_ptr_vector_set(p_discovered_vec, lid, NULL);

	/* we if are in the first sweep and in reassign lids mode
	   we should ignore all the available info and simply define one
	   huge empty range */
	if (p_mgr->p_subn->first_time_master_sweep == TRUE &&
	    p_mgr->p_subn->opt.reassign_lids == TRUE) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Skipping all lids as we are reassigning them\n");
		p_range = calloc(1, sizeof(osm_lid_mgr_range_t));
		if (p_range)
			p_range->min_lid = 1;
		goto AfterScanningLids;
	}

	/* go over all discovered ports and mark their entries */
	p_port_guid_tbl = &p_mgr->p_subn->port_guid_tbl;

	for (p_port = (osm_port_t *) cl_qmap_head(p_port_guid_tbl);
	     p_port != (osm_port_t *) cl_qmap_end(p_port_guid_tbl);
	     p_port = (osm_port_t *) cl_qmap_next(&p_port->map_item)) {
		osm_port_get_lid_range_ho(p_port, &disc_min_lid, &disc_max_lid);
		disc_min_lid = trim_lid(disc_min_lid);
		if ((p_mgr->p_subn->opt.lmc &&
		    (!p_port->p_node->sw ||
		     osm_switch_sp0_is_lmc_capable(p_port->p_node->sw,
				p_mgr->p_subn))) ||
		    (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_ROUTER &&
				p_mgr->p_subn->opt.rtr_lmc))
			num_lids = lmc_num_lids;
		else
			num_lids = 1;

		disc_max_lid = disc_min_lid + num_lids - 1;

		/*
		 * If forcing SM LID, and found port with conflicting LID, dont
		 * update it in LID-to-port table.
		 */
		if (sm_min_lid &&
		    osm_port_get_guid(p_port) != sm_port_guid &&
		    ((disc_min_lid <= sm_min_lid && sm_min_lid <= disc_max_lid) ||
		     (sm_min_lid <= disc_min_lid && disc_min_lid <= sm_max_lid)))
			continue;

		if (sm_min_lid && osm_port_get_guid(p_port) == sm_port_guid)
			continue;

		/*
		 * If the FLID feature is enabled, and found port with conflicting LID, don't
		 * update it in LID-to-port table.
		 */
		if (osm_subn_is_flids_enabled(p_mgr->p_subn) &&
		    ((disc_min_lid <= global_flid_start &&
		       global_flid_start <= disc_max_lid) ||
		     (global_flid_start <= disc_min_lid &&
		       disc_min_lid <= global_flid_end))) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"discovered LID (%u-%u) of port guid: 0x%016" PRIx64
				" conflicts with global FLID range (%u-%u)\n",
				disc_min_lid, disc_max_lid,
				cl_ntoh64(osm_port_get_guid(p_port)),
				global_flid_start,
				global_flid_end);
			continue;
		}

		if (disc_min_lid && ((num_lids == 1) ||
		    (disc_min_lid & lmc_mask) == disc_min_lid)) {
			for (lid = disc_min_lid; lid <= disc_max_lid; lid++) {
				cl_ptr_vector_set(p_discovered_vec, lid, p_port);
				p_mgr->used_lids[lid] |= OSM_RESERVED_LID_DISCOVERED;
			}
		}

		/*
		 * If virtual port has LID, but does not have persistent LID,
		 * mark current lids as used.
		 */
		top_index = osm_port_get_vport_top_index(p_port);
		for (index = 1 ; index <= top_index ; index++) {
			p_vport = osm_port_get_vport(p_port, index);
			if (p_vport &&
			    (lid = trim_lid(p_vport->vport_info.vport_lid)) &&
			    osm_db_guid2lid_get(p_mgr->p_g2l,
						cl_ntoh64(osm_vport_get_guid(p_vport)),
						NULL, NULL)) {
				p_mgr->used_lids[lid] |= OSM_RESERVED_LID_DISCOVERED;
			}
		}

		/*
		 * Check if port has guid2lid entry. If not, check that currenlty
		 * used LID is free and matches LMC, and set p_mgr->used_lids
		 * accordingly.
		 */
		if (osm_db_guid2lid_get(p_mgr->p_g2l,
					cl_ntoh64(osm_port_get_guid(p_port)),
					&db_min_lid, &db_max_lid)) {
			if (!disc_min_lid || disc_max_lid > IB_LID_UCAST_END_HO)
				continue;

			/* Single LID, just mark LID as used */
			if (num_lids == 1) {
				p_mgr->used_lids[disc_min_lid] |= OSM_RESERVED_LID_DISCOVERED;
				continue;
			}

			/* Chcek if base LID matches LMC */
			if ((disc_min_lid & lmc_mask) != disc_min_lid)
				continue;

			/*
			 * Check that LIDs range is free.
			 * If free, it can be used by current device, hence, mark as used.
			 */
			is_free = TRUE;
			for (lid = disc_min_lid; lid <= disc_max_lid; lid++) {
				if (p_mgr->used_lids[lid]) {
					is_free = FALSE;
					break;
				}
			}
			if (!is_free)
				continue;

			/* Mark LID range as used */
			for (lid = disc_min_lid; lid <= disc_max_lid; lid++)
				p_mgr->used_lids[lid] |= OSM_RESERVED_LID_DISCOVERED;

			continue;
		}

		/*
		 * Make sure the guid2lid entry is valid. If not, clean it.
		 * guid2lid is not valid only if base LID is not aligned to LMC
		 * or LID range is not wide enough to match LMC.
		 */
		if (num_lids > 1 &&
		    ((db_min_lid & lmc_mask) != db_min_lid ||
		     db_max_lid - db_min_lid + 1 < num_lids)) {
			/* Not aligned, or not wide enough, then remove the entry */
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"Cleaning persistent entry for guid:"
				"0x%016" PRIx64 " illegal range:[0x%x:0x%x]\n",
				cl_ntoh64(osm_port_get_guid(p_port)),
				db_min_lid, db_max_lid);
			osm_db_guid2lid_delete(p_mgr->p_g2l,
					       cl_ntoh64
					       (osm_port_get_guid(p_port)));
			for (lid = db_min_lid; lid <= db_max_lid; lid++)
				p_mgr->used_lids[lid] &= OSM_RESERVED_LID_DISCOVERED;
		}
	}

	/* Make sure the defined sm lid is marked as discovered */
	if (sm_min_lid) {
		for (lid = sm_min_lid; lid <= sm_max_lid; lid++) {
			p_mgr->used_lids[lid] |= OSM_RESERVED_LID_DISCOVERED;
		}
	}

	/*
	 * Construct free LIDs ranages according to p_mgr->used_lids.
	 */
	for (lid = 1; lid <= IB_LID_UCAST_END_HO; lid++) {
		if (p_mgr->used_lids[lid] == 0) {
			if (p_range)
				p_range->max_lid = lid;
			else {
				p_range = calloc(1, sizeof(osm_lid_mgr_range_t));
				if (p_range) {
					p_range->min_lid = lid;
					p_range->max_lid = lid;
				}
			}
		/* this lid is used so we need to finalize the previous free range */
		} else if (p_range) {
			cl_qlist_insert_tail(&p_mgr->free_ranges,
					     &p_range->item);
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"new free lid range [%u:%u]\n",
				p_range->min_lid, p_range->max_lid);
			p_range = NULL;
		}
	}

AfterScanningLids:
	/* after scanning all known lids we need to extend the last range
	   to the max allowed lid */
	if (p_range) {
		p_range->max_lid = p_mgr->p_subn->max_ucast_lid_ho;
		cl_qlist_insert_tail(&p_mgr->free_ranges, &p_range->item);
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"final free lid range [%u:%u]\n",
			p_range->min_lid, p_range->max_lid);
	}

	OSM_LOG_EXIT(p_mgr->p_log);
	return status;
}

/**********************************************************************
 check if the given range of lids is free
**********************************************************************/
static boolean_t lid_mgr_is_range_not_persistent(IN osm_lid_mgr_t * p_mgr,
						 IN uint16_t lid,
						 IN uint16_t num_lids)
{
	uint16_t i;

	for (i = lid; i < lid + num_lids; i++)
		if (p_mgr->used_lids[i] & OSM_RESERVED_LID_PERSISTENT)
			return FALSE;

	return TRUE;
}

/*
 * Get min and max values and create free range for these values
 * If p_prev_item is given, insert the new created range after it
 * If p_prev_item is not given, insert new range at tail
 *
 * return pointer for new range or NULL if failed to allocate memory
 */
static osm_lid_mgr_range_t *create_free_range(osm_lid_mgr_t *p_mgr,
					      uint16_t min, uint16_t max,
					      cl_list_item_t *p_prev_item)
{
	osm_lid_mgr_range_t *p_range;

	p_range = calloc(1, sizeof(osm_lid_mgr_range_t));
	if (p_range) {
		p_range->min_lid = min;
		p_range->max_lid = max;
		if (p_prev_item)
			cl_qlist_insert_next(&p_mgr->free_ranges, p_prev_item,
					     &p_range->item);
		else
			cl_qlist_insert_tail(&p_mgr->free_ranges, &p_range->item);

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Created new free lid range [%u:%u]\n", min, max);
	} else
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 031D: "
			"Failed to allocate new free range\n");

	return p_range;
}

/*
 * Clear all unused lids (lids which are mapped to physical or virtual ports
 * which are currently down)
 * Return the number of lids which were cleared
 */
static uint16_t osm_lid_mgr_clear_unused_lids(osm_lid_mgr_t *p_mgr)
{
	osm_port_t *p_port;
	cl_qlist_t guids;
	osm_db_guid_elem_t *p_item;
	uint16_t lid, min_lid, max_lid, cleared = 0;
	cl_list_item_t *p_range_item;
	cl_qlist_init(&guids);

	if (osm_db_guid2lid_guids(p_mgr->p_g2l, &guids)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 031E: "
			"could not get guid list\n");
		goto EXIT;
	}

	while ((p_item = (osm_db_guid_elem_t *) cl_qlist_remove_head(&guids))
	       != (osm_db_guid_elem_t *) cl_qlist_end(&guids)) {
		if (osm_db_guid2lid_get(p_mgr->p_g2l, p_item->guid,
					&min_lid, &max_lid)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 031F: "
				"could not get lid for guid:0x%016" PRIx64 "\n",
				p_item->guid);
			free(p_item);
			continue;
		}

		p_port = osm_get_port_by_lid_ho(p_mgr->p_subn, min_lid);
		if (!p_port) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Clearing "
				"unused lid range [%u:%u] (mapped for guid 0x%"
				PRIx64 ") from guid2lid db\n",
				min_lid, max_lid, p_item->guid);
			osm_db_guid2lid_delete(p_mgr->p_g2l, p_item->guid);
			for (lid = min_lid ; lid <= max_lid ; lid++)
				p_mgr->used_lids[lid] = 0;
			cleared++;
		}
		free(p_item);
	}

	if (!cleared)
		goto EXIT;

	/* We need to rebuild the empty ranges list */

	/* Cleanup current list */
	while ((p_range_item = cl_qlist_remove_head(&p_mgr->free_ranges)) !=
	       cl_qlist_end(&p_mgr->free_ranges))
		free((osm_lid_mgr_range_t *) p_range_item);

	/* Build new list */
	min_lid = 0;
	for (lid = 1 ; lid <= IB_LID_UCAST_END_HO ; lid++) {
		if (p_mgr->used_lids[lid]) {
			if (min_lid) {
				if (!create_free_range(p_mgr, min_lid, lid -1,
						       NULL)) {
					cleared = 0;
					goto EXIT;
				}
				min_lid = 0;
			}
		} else if (lid == IB_LID_UCAST_END_HO) {
			if (!min_lid)
				min_lid = lid;

			if (!create_free_range(p_mgr, min_lid, lid, NULL))
					cleared = 0;
		} else if (!min_lid)
			min_lid = lid;
	}
EXIT:
	return cleared;
}

static uint16_t osm_lid_mgr_clear_unused_flids(osm_lid_mgr_t *p_mgr)
{
	osm_port_t *p_port;
	cl_qlist_t guids;
	osm_db_guid_elem_t *p_item;
	uint16_t flid, cleared = 0;

	cl_qlist_init(&guids);

	if (osm_db_guid2lid_guids(p_mgr->p_g2fl, &guids)) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0323: "
			"could not get guid list\n");
		goto EXIT;
	}

	while ((p_item = (osm_db_guid_elem_t *) cl_qlist_remove_head(&guids))
	       != (osm_db_guid_elem_t *) cl_qlist_end(&guids)) {
		if (osm_db_guid2flid_get(p_mgr->p_g2fl, p_item->guid, &flid)) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0324: "
				"could not get FLID for guid:0x%016" PRIx64 "\n",
				p_item->guid);
			free(p_item);
			continue;
		}

		p_port = osm_get_port_by_guid(p_mgr->p_subn, cl_hton64(p_item->guid));
		if (!p_port) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Clearing "
				"unused FLID %u (mapped for guid 0x%" PRIx64 ") "
				"from guid2flid db\n",
				flid, p_item->guid);
			osm_db_guid2flid_delete(p_mgr->p_g2fl, p_item->guid);
			p_mgr->used_lids[flid] = 0;
			cleared++;
		}
		free(p_item);
	}

	if (cleared) {
		lid_mgr_reset_flid_free_range(p_mgr);
		/*
		 * after successful build of free ranges mark all global flid range,
		 * so that cleared in process used[flid] entries
		 * are not assigned to unicast lids.
		 */
		lid_mgr_mark_global_flids(p_mgr);
	}
EXIT:
	return cleared;
}

typedef enum lid_mgr_lid_type_enum {
        LID = 0,
        FLID
} lid_mgr_lid_type_enum;

static int lid_mgr_find_free_range(IN osm_lid_mgr_t * p_mgr,
				   IN uint8_t num_lids,
				   OUT uint16_t * p_min_lid,
				   OUT uint16_t * p_max_lid,
				   lid_mgr_lid_type_enum lid_type) {
	cl_list_item_t *p_item;
	cl_list_item_t *p_next_item;
	osm_lid_mgr_range_t *p_range = NULL;
	uint8_t max_lmc;
	uint8_t lmc_num_lids;
	uint16_t lmc_mask;
	uint16_t lid;
	cl_qlist_t *p_lid_ranges;

	max_lmc = osm_subn_get_max_lmc(p_mgr->p_subn);
	lmc_num_lids = (1 << max_lmc);
	lmc_mask = ~(lmc_num_lids - 1);

	if (lid_type == LID) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "MAX LMC = %u, number LIDs = %u\n",
			max_lmc, num_lids);
		p_lid_ranges = &p_mgr->free_ranges;
	} else {
		p_lid_ranges = &p_mgr->free_flid_ranges;
	}
	/*
	   Search the list of free lid ranges for a range which is big enough
	 */
	p_item = cl_qlist_head(p_lid_ranges);
	while (p_item != cl_qlist_end(p_lid_ranges)) {
		p_next_item = cl_qlist_next(p_item);
		p_range = (osm_lid_mgr_range_t *) p_item;

		lid = p_range->min_lid;

		/*
		 * If we require more then one lid we must align to LMC.
		 * This is relevant only for 'LID' calculation.
		 */
		if (num_lids > 1) {
			if ((lid & lmc_mask) != lid)
				lid = (lid + lmc_num_lids) & lmc_mask;
		}

		/* but we can be out of the range */
		if (lid + num_lids - 1 <= p_range->max_lid) {
			/* ok let us use that range */
			if (lid + num_lids - 1 == p_range->max_lid) {
				/* we consumed the entire range */
				cl_qlist_remove_item(p_lid_ranges, p_item);
				free(p_item);
			} else
				/* only update the available range */
				p_range->min_lid = lid + num_lids;

			*p_min_lid = lid;
			*p_max_lid = (uint16_t) (lid + num_lids - 1);
			if (lid_type == FLID) {
				/* update the max assigned FLID */
				p_mgr->max_flid_assigned = lid;
			}
			return 0;
		}
		p_item = p_next_item;
	}
	/* didn't find free range */
	return 1;
}

/**********************************************************************
find a free lid range
**********************************************************************/
static void lid_mgr_find_free_lid_range(IN osm_lid_mgr_t * p_mgr,
					IN uint8_t num_lids,
					OUT uint16_t * p_min_lid,
					OUT uint16_t * p_max_lid)
{
	uint16_t cleared_lids;

	if (!lid_mgr_find_free_range(p_mgr, num_lids, p_min_lid, p_max_lid, LID))
		return;
	else {
		/* Couldn't find a free range of lids. */
		cleared_lids = osm_lid_mgr_clear_unused_lids(p_mgr);
		if (cleared_lids >= num_lids) {
			/* found free range after clearing unused lids */
			if (!lid_mgr_find_free_range(p_mgr, num_lids,
						     p_min_lid, p_max_lid, LID))
				return;
		}
	}

	*p_min_lid = *p_max_lid = 0;
	/* if we run out of lids, give an error and abort! */
	OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0307: "
		"OPENSM RAN OUT OF LIDS!!!\n");
	OSM_ASSERT(0);
}

static void lid_mgr_find_free_flid(IN osm_lid_mgr_t * p_mgr, OUT uint16_t * p_flid)
{
	uint16_t cleared_lids;

	if (!lid_mgr_find_free_range(p_mgr, 1, p_flid, p_flid, FLID))
		return;
	else {
		/* Couldn't find a free flid range of lids. */
		cleared_lids = osm_lid_mgr_clear_unused_flids(p_mgr);
		if (cleared_lids) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"%u unused FLIDs cleared from guid2flid db.\n", cleared_lids);

			/* found free range after clearing unused lids */
			if (!lid_mgr_find_free_range(p_mgr, 1, p_flid, p_flid, FLID))
				return;
		}
	}

	*p_flid = 0;
	OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0309: SM Ran out of FLIDs.\n");
}

static void lid_mgr_cleanup_discovered_port_lid_range(IN osm_lid_mgr_t * p_mgr,
						      IN osm_port_t * p_port)
{
	cl_ptr_vector_t *p_discovered_vec = &p_mgr->p_subn->port_lid_tbl;
	uint16_t lid, min_lid, max_lid;
	uint16_t max_tbl_lid =
	    (uint16_t) (cl_ptr_vector_get_size(p_discovered_vec));

	osm_port_get_lid_range_ho(p_port, &min_lid, &max_lid);
	min_lid = trim_lid(min_lid);
	max_lid = trim_lid(max_lid);
	for (lid = min_lid; lid <= max_lid; lid++)
		if (lid < max_tbl_lid &&
		    p_port == cl_ptr_vector_get(p_discovered_vec, lid))
			cl_ptr_vector_set(p_discovered_vec, lid, NULL);
}

/*
 * This function gets lid range and remove this range from the free ranges
 * list of the lid manager
 */
static int lid_mgr_set_lids_not_free(osm_lid_mgr_t *p_mgr, uint16_t min_lid,
				     uint16_t max_lid)
{
	cl_list_item_t *p_item;
	cl_qlist_t *p_list = &p_mgr->free_ranges;
	boolean_t found = FALSE;
	osm_lid_mgr_range_t *p_range = NULL;

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Removing range [%u-%u] from the "
		"free ranges list\n", min_lid, max_lid);

	/*
	 * The range we would like to remove from free ranges MUST be included
	 * in a single range item and the free ranges list.
	 * First we need to jump over all the ranges in the list which are
	 * completely lower than the range we would like to remove and get to
	 * the free range that should include our range.
	 */
	p_item = cl_qlist_head(p_list);
	while ((p_item != cl_qlist_end(p_list)) && !found) {
		p_range = (osm_lid_mgr_range_t*)p_item;

		if (min_lid > p_range->max_lid)
			p_item = cl_qlist_next(p_item);
		else
			found = TRUE;
	}

	/*
	 * If the required range was not found in the free ranges it means
	 * that we had some kind of incoherency in our db so we print an error
	 * message, however since it is not in the free ranges db and we
	 * probably just assigned the lid, there is nothing else to do as
	 * the db is now probably aligned
	 */
	if (!found) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 031B: "
			"Range [%u-%u] not found in free ranges list\n",
			min_lid, max_lid);
		return 1;
	}

	/*
	 * Our range must be part of the current free range
	 * If it doesn't it means that our range was not part of the
	 * free ranges to begin with
	 */
	if (min_lid < p_range->min_lid || max_lid > p_range->max_lid) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 031C: "
			"Failed to remove range [%u-%u] from free "
			"ranges, this range is not part of free ranges"
			", Found at free range [%u-%u]\n",
			min_lid, max_lid, p_range->min_lid,
			p_range->max_lid);
		return 1;
	}

	/*
	 * Handle the specific case that our range and the current free
	 * range are identical (just remove this free range)
	 */
	if (min_lid == p_range->min_lid && max_lid == p_range->max_lid) {
		cl_qlist_remove_item(p_list, p_item);
		free(p_item);
		return 0;
	}

	/*
	 * We confirmed that our range is inside the current free range,
	 * now there are three options:
	 *
	 *	- Our range at the beginning of the free range:
	 *		Just need to update a new min to the free range
	 *
	 *	- Our range at the end of the free range:
	 *		Just need to update a new max to the free range
	 *
	 *	- Our range is in the middle of the free range:
	 *		Need to split the free range into two ranges
	 */
	if (p_range->min_lid == min_lid) {
		/* Our range at the beginning of the free range */
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Updated range [%u-%u] "
			"to [%u-%u]\n", p_range->min_lid, p_range->max_lid,
			max_lid + 1, p_range->max_lid);
		p_range->min_lid = max_lid + 1;
	} else if (p_range->max_lid == max_lid) {
		/* Our range at the end of the free range */
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Updated range [%u-%u] "
			"to [%u-%u]\n", p_range->min_lid, p_range->max_lid,
			p_range->min_lid, min_lid - 1);
		p_range->max_lid = min_lid - 1;
	} else {
		/* Our range is in the middle of the free range */
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Splitting range [%u-%u] "
			"to ranges [%u-%u] and [%u-%u]\n",
			p_range->min_lid, p_range->max_lid, p_range->min_lid,
			min_lid - 1, max_lid + 1, p_range->max_lid);

		/* First create new free range for the part of the free range
		 * which is higher than our range
		 */
		if (!create_free_range(p_mgr, max_lid + 1, p_range->max_lid,
				       (cl_list_item_t *)p_range))
			return 1;

		/* And update current range for the part of the range which
		 * is lower than our range
		 */
		p_range->max_lid = min_lid - 1;
	}

	return 0;
}

/**********************************************************************
 0.1 if the port info lid matches the guid2lid return 0
 0.2 if the port info has a lid and that range is empty in
     port_lid_tbl, return 0 and update the port_lid_tbl and
     guid2lid
 0.3 else find an empty space in port_lid_tbl, update the
 port_lid_tbl and guid2lid, return 1 to flag a change required.
**********************************************************************/
static int lid_mgr_get_port_lid(IN osm_lid_mgr_t * p_mgr,
				IN osm_port_t * p_port,
				IN osm_vport_t * p_vport,
				IN uint16_t req_lid,
				OUT uint16_t * p_min_lid,
				OUT uint16_t * p_max_lid)
{
	uint16_t lid, min_lid, max_lid, disc_base_lid;
	uint64_t guid;
	uint8_t num_lids;
	int lid_changed = 0;
	uint16_t lmc_mask;

	OSM_LOG_ENTER(p_mgr->p_log);

	/* get the lid from the guid2lid */
	if (p_vport) {
		guid = cl_ntoh64(osm_vport_get_guid(p_vport));
		disc_base_lid = cl_ntoh16(osm_vport_get_lid(p_vport));
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Getting lid for virtual "
			"port 0x%016" PRIx64 "\n", guid);
	} else {
		guid = cl_ntoh64(osm_port_get_guid(p_port));
		disc_base_lid = cl_ntoh16(osm_port_get_base_lid(p_port));
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Getting lid for port "
			"0x%016" PRIx64 "\n",
			guid);
	}
	/* if the port is a base switch port 0 then we only need one lid */
	if (!p_vport &&
	    ((p_mgr->p_subn->opt.lmc &&
	     (!p_port->p_node->sw ||
	     osm_switch_sp0_is_lmc_capable(p_port->p_node->sw,
					p_mgr->p_subn))) ||
	    (osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_ROUTER &&
			p_mgr->p_subn->opt.rtr_lmc)))
		num_lids = (1 << osm_subn_get_max_lmc(p_mgr->p_subn));
	else
		num_lids = 1;

	if (p_mgr->p_subn->first_time_master_sweep == TRUE &&
	    p_mgr->p_subn->opt.reassign_lids == TRUE)
		goto AssignLid;

	lmc_mask = ~(num_lids - 1);

	/* if the port matches the guid2lid */
	if (!osm_db_guid2lid_get(p_mgr->p_g2l, guid, &min_lid, &max_lid)) {
		*p_min_lid = min_lid;
		*p_max_lid = min_lid + num_lids - 1;
		if (min_lid == disc_base_lid)
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "0x%016" PRIx64
				" matches its known lid:%u\n", guid, min_lid);
		else {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"0x%016" PRIx64 " with lid:%u "
				"does not match its known lid:%u\n",
				guid, disc_base_lid, min_lid);
			if (!p_vport)
				lid_mgr_cleanup_discovered_port_lid_range(p_mgr,
									  p_port);
			/* we still need to send the setting to the target port */
			lid_changed = 1;
		}
		goto NewLidSet;
	} else
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"0x%016" PRIx64 " has no persistent lid assigned\n",
			guid);

	if (req_lid && disc_base_lid != req_lid) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"port 0x%016" PRIx64 " lid %u does not match "
			"requested lid %u\n",
			guid, disc_base_lid, req_lid);
		goto AssignLid;
	}

	/* if the port info carries a lid it must be lmc aligned and not mapped
	   by the pesistent storage  */
	min_lid = disc_base_lid;

	/* we want to ignore the discovered lid if we are also on first sweep of
	   reassign lids flow */
	if (min_lid) {
		/* make sure lid is valid */
		if ((min_lid & lmc_mask) == min_lid) {
			/* is it free */
			if (lid_mgr_is_range_not_persistent
			    (p_mgr, min_lid, num_lids)) {
				*p_min_lid = min_lid;
				*p_max_lid = min_lid + num_lids - 1;
				OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
					"0x%016" PRIx64
					" lid range:[%u-%u] is free\n",
					guid, *p_min_lid, *p_max_lid);
				goto NewLidSet;
			} else
				OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
					"0x%016" PRIx64 " existing lid "
					"range:[%u:%u] is not free\n",
					guid, min_lid, min_lid + num_lids - 1);
		} else
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"0x%016" PRIx64 " existing lid range:"
				"[%u:%u] is not lmc aligned\n",
				guid, min_lid, min_lid + num_lids - 1);
	}

AssignLid:
	/* first cleanup the existing discovered lid range */
	if (!p_vport)
		lid_mgr_cleanup_discovered_port_lid_range(p_mgr, p_port);

	if (req_lid) {
		*p_min_lid = req_lid;
		*p_max_lid = req_lid + num_lids - 1;
	} else {
		/* find an empty space */
		lid_mgr_find_free_lid_range(p_mgr, num_lids, p_min_lid, p_max_lid);
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"0x%016" PRIx64 " assigned a new lid range:[%u-%u]\n",
			guid, *p_min_lid, *p_max_lid);
	}

	lid_changed = 1;

	/* Avoid infinite loop in case of no free lids */
	if (*p_min_lid == 0)
		goto EXIT;

NewLidSet:
	if (*p_min_lid == 0)
                goto AssignLid;

	/*
	 * Update free ranges
	 * Required when called from LID manager
	 */
	if (p_vport && !lid_changed && !p_mgr->used_lids[*p_min_lid]) {
		lid_mgr_set_lids_not_free(p_mgr, *p_min_lid, *p_max_lid);
	}

	/* update the guid2lid db and used_lids */
	osm_db_guid2lid_set(p_mgr->p_g2l, guid, *p_min_lid, *p_max_lid);
	for (lid = *p_min_lid; lid <= *p_max_lid; lid++)
		p_mgr->used_lids[lid] = OSM_RESERVED_LID_PERSISTENT | OSM_RESERVED_LID_DISCOVERED;

	/* make sure the assigned lids are marked in port_lid_tbl */
	if (!p_vport)
		for (lid = *p_min_lid; lid <= *p_max_lid; lid++)
			cl_ptr_vector_set(&p_mgr->p_subn->port_lid_tbl, lid, p_port);

EXIT:
	OSM_LOG_EXIT(p_mgr->p_log);
	return lid_changed;
}

uint16_t osm_lid_mgr_get_switch_flid(IN osm_lid_mgr_t * p_mgr,
				     IN osm_port_t * p_port)
{
	uint16_t flid = 0;
	uint64_t guid;

	OSM_LOG_ENTER(p_mgr->p_log);

	guid = cl_ntoh64(osm_port_get_guid(p_port));
	
	/* try to assign new flid if not in db */
	if (osm_db_guid2flid_get(p_mgr->p_g2fl, guid, &flid)) {
		lid_mgr_find_free_flid(p_mgr, &flid);

		if (flid) {
			/* update the guid2lid db and used_lids */
			osm_db_guid2flid_set(p_mgr->p_g2fl, guid, flid);
			/*
			 * Currently all global FLIDs are marked as persistent,
			 * mark that the flid is now also assigned (discovered).
			 */
			p_mgr->used_lids[flid] =
				OSM_RESERVED_LID_PERSISTENT | OSM_RESERVED_LID_DISCOVERED;
		}
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Assigned port guid 0x%016" PRIx64 " with flid %u.\n",
			cl_ntoh64(osm_port_get_guid(p_port)), flid);

	}

	OSM_LOG_EXIT(p_mgr->p_log);
	return flid;
}

static int lid_mgr_set_physp_pi(IN osm_lid_mgr_t * p_mgr,
				IN osm_port_t * p_port,
				IN osm_physp_t * p_physp, IN ib_net16_t lid)
{
	uint8_t payload[IB_SMP_DATA_SIZE];
	ib_port_info_t *p_pi = (ib_port_info_t *) payload;
	const ib_port_info_t *p_old_pi;
	osm_madw_context_t context;
	osm_node_t *p_node;
	ib_api_status_t status;
	uint8_t port_num;
	boolean_t send_set = FALSE;
	boolean_t update_mkey = FALSE;
	boolean_t send_client_rereg = FALSE;
	int ret = 0;

	OSM_LOG_ENTER(p_mgr->p_log);

	/*
	   Don't bother doing anything if this Physical Port is not valid.
	   This allows simplified code in the caller.
	 */
	if (!p_physp)
		goto Exit;

	port_num = osm_physp_get_port_num(p_physp);
	p_node = osm_physp_get_node_ptr(p_physp);

	if (osm_node_get_type(p_node) == IB_NODE_TYPE_SWITCH && port_num != 0) {
		/*
		   Switch ports that are not numbered 0 should not be set
		   with the following attributes as they are set later
		   (during NO_CHANGE state in link mgr).
		 */
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Skipping switch port %u, GUID 0x%016" PRIx64 "\n",
			port_num, cl_ntoh64(osm_physp_get_port_guid(p_physp)));
		goto Exit;
	}

	p_old_pi = &p_physp->port_info;

	/*
	   First, copy existing parameters from the PortInfo attribute we
	   already have for this node.

	   Second, update with default values that we know must be set for
	   every Physical Port and the LID and set the neighbor MTU field
	   appropriately.

	   Third, send the SMP to this physical port.
	 */

	memcpy(payload, p_old_pi, sizeof(ib_port_info_t));

	/*
	   Should never write back a value that is bigger then 3 in
	   the PortPhysicalState field, so cannot simply copy!

	   Actually we want to write there:
	   port physical state - no change
	   link down default state = polling
	   port state - no change
	 */
	p_pi->state_info2 = 0x02;
	ib_port_info_set_port_state(p_pi, IB_LINK_NO_CHANGE);

	if (ib_port_info_get_link_down_def_state(p_pi) !=
	    ib_port_info_get_link_down_def_state(p_old_pi))
		send_set = TRUE;

	/* didn't get PortInfo before */
	if (!ib_port_info_get_port_state(p_old_pi))
		send_set = TRUE;

	p_pi->m_key = osm_physp_calc_mkey(p_mgr->p_subn, p_physp);
	if (memcmp(&p_pi->m_key, &p_old_pi->m_key, sizeof(p_pi->m_key))) {
		update_mkey = TRUE;
		send_set = TRUE;
	}

	/*
	 * Make sure that only subnet prefix part of gid prefix is set,
	 * don't change the flid part.
	 * This if should happen only first time the configured subnet prefix is not set yet.
	 */
	if (!is_equal_subnet_prefix(p_old_pi->subnet_prefix,
				    p_mgr->p_subn->opt.subnet_prefix)) {
		p_pi->subnet_prefix =
			(p_old_pi->subnet_prefix & (~IB_SITE_LOCAL_VALID_BITS_SUBNET_MASK)) |
			(p_mgr->p_subn->opt.subnet_prefix & IB_SITE_LOCAL_VALID_BITS_SUBNET_MASK);

		send_set = TRUE;
	}

	p_port->lid = lid;
	p_pi->base_lid = lid;
	if (memcmp(&p_pi->base_lid, &p_old_pi->base_lid,
		   sizeof(p_pi->base_lid))) {
		/*
		 * Reset stored base_lid.
		 * On successful send, we'll update it when we'll get a reply.
		 */
		osm_physp_set_base_lid(p_physp, 0);
		send_set = TRUE;
		p_mgr->dirty = TRUE;
	}

	/*
	   We are updating the ports with our local sm_base_lid
	   if for some reason currently received SM LID is different from our SM LID,
	   need to send client reregister to this port
	*/
	p_pi->master_sm_base_lid = p_mgr->p_subn->sm_base_lid;
	if (memcmp(&p_pi->master_sm_base_lid, &p_old_pi->master_sm_base_lid,
		   sizeof(p_pi->master_sm_base_lid))) {
		send_client_rereg = TRUE;
		send_set = TRUE;
	}

	p_pi->m_key_lease_period = p_mgr->p_subn->opt.m_key_lease_period;
	if (memcmp(&p_pi->m_key_lease_period, &p_old_pi->m_key_lease_period,
		   sizeof(p_pi->m_key_lease_period)))
		send_set = TRUE;

	p_pi->mkey_lmc = 0;
	ib_port_info_set_mpb(p_pi, p_mgr->p_subn->opt.m_key_protect_bits);
	if (ib_port_info_get_mpb(p_pi) != ib_port_info_get_mpb(p_old_pi))
		send_set = TRUE;

	/*
	   we want to set the timeout for both the switch port 0
	   and the CA ports
	 */
	ib_port_info_set_timeout(p_pi, p_mgr->p_subn->opt.subnet_timeout);
	if (ib_port_info_get_timeout(p_pi) !=
	    ib_port_info_get_timeout(p_old_pi))
		send_set = TRUE;

	if (port_num != 0) {
		/*
		   CAs don't have a port 0, and for switch port 0,
		   the state bits are ignored.
		   This is not the switch management port
		 */

		/* p_pi->mkey_lmc is initialized earlier */
		uint8_t set_lmc = 0;
		if (p_mgr->p_subn->opt.lmc)
			set_lmc = p_mgr->p_subn->opt.lmc;
		else if (osm_node_get_type(p_node) == IB_NODE_TYPE_ROUTER)
			set_lmc = p_mgr->p_subn->opt.rtr_lmc;

		ib_port_info_set_lmc(p_pi, set_lmc);
		p_physp->expected_lmc = ib_port_info_get_lmc(p_pi);
		if (ib_port_info_get_lmc(p_pi) !=
		    ib_port_info_get_lmc(p_old_pi)) {
			send_set = TRUE;
			p_mgr->dirty = TRUE;
		}

		/*
		   Several timeout mechanisms:
		 */
		ib_port_info_set_phy_and_overrun_err_thd(p_pi,
							 p_mgr->p_subn->opt.
							 local_phy_errors_threshold,
							 p_mgr->p_subn->opt.
							 overrun_errors_threshold);

		if (p_pi->error_threshold != p_old_pi->error_threshold)
			send_set = TRUE;

	} else if (ib_switch_info_is_enhanced_port0(&p_node->sw->switch_info)) {
		/*
		 * Configure Enh. SP0:
		 * Set MTU according to the mtu_cap.
		 * Set LMC if lmc_esp0 is defined.
		 */
		ib_port_info_set_neighbor_mtu(p_pi,
					      ib_port_info_get_mtu_cap
					      (p_old_pi));
		if (ib_port_info_get_neighbor_mtu(p_pi) !=
		    ib_port_info_get_neighbor_mtu(p_old_pi))
			send_set = TRUE;

		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Updating neighbor_mtu on switch GUID 0x%016" PRIx64
			" port 0 to:%u\n",
			cl_ntoh64(osm_physp_get_port_guid(p_physp)),
			ib_port_info_get_neighbor_mtu(p_pi));

		/* Configure LMC on enhanced SP0 */
		if (p_mgr->p_subn->opt.lmc_esp0) {
			/* p_pi->mkey_lmc is initialized earlier */
			ib_port_info_set_lmc(p_pi, p_mgr->p_subn->opt.lmc);
			p_physp->expected_lmc = ib_port_info_get_lmc(p_pi);
			if (ib_port_info_get_lmc(p_pi) !=
			    ib_port_info_get_lmc(p_old_pi))
				send_set = TRUE;
		}
	}

	memset(&context, 0, sizeof(osm_madw_context_t));
	context.pi_context.node_guid = osm_node_get_node_guid(p_node);
	context.pi_context.port_guid = osm_physp_get_port_guid(p_physp);
	context.pi_context.set_method = TRUE;
	context.pi_context.light_sweep = FALSE;
	context.pi_context.active_transition = FALSE;

	/*
	   For ports supporting the ClientReregistration Vol1 (v1.2) p811 14.4.11:
	   need to set the cli_rereg bit when current SM LID at the Host
	   is different from our SM LID, if we are in first_time_master_sweep
	   if this port was just now discovered, then we should also set
	   the cli_rereg bit. We know that the port was just discovered if its
	   is_new field is set.
	   Also in case of timeout for previous PortInfo with client_reregister
	   bit set, in current PortInfo it will be set again(clien_rereg flag).
	   Moreover, we won't set client_reregister for a port that was discovered
	   only through an ignored port or to switch that has ignored ports.
	 */
	if (p_port->disc_through != OSM_DISC_IGNORED &&
	    (send_client_rereg ||
	     p_mgr->p_subn->first_time_master_sweep == TRUE || p_port->is_new)
	    && p_mgr->p_subn->opt.client_rereg_mode
	    && (p_old_pi->capability_mask & IB_PORT_CAP_HAS_CLIENT_REREG)) {
		/* Mark port for sending client reregister */
		p_port->client_rereg = TRUE;
	}

	if (p_port->client_rereg &&
	    p_mgr->p_subn->opt.client_rereg_mode == OSM_CLIENT_REREG_MODE_LID_MGR) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
			"Setting client rereg on %s, port %d\n",
			p_port->p_node->print_desc, p_port->p_physp->port_num);
		ib_port_info_set_client_rereg(p_pi, 1);
		context.pi_context.client_rereg = TRUE;
		send_set = TRUE;
	} else {
		ib_port_info_set_client_rereg(p_pi, 0);
		context.pi_context.client_rereg = FALSE;
	}

	/* We need to send the PortInfo Set request with the new sm_lid
	   in the following cases:
	   1. There is a change in the values (send_set == TRUE)
	   2. first_time_master_sweep flag on the subnet is TRUE. This means the
	   SM just became master, and it then needs to send a PortInfo Set to
	   every port.
	 */
	if (p_mgr->p_subn->first_time_master_sweep == TRUE)
		send_set = TRUE;

	if (!send_set)
		goto Exit;

	status = osm_req_set_pi(p_mgr->sm, p_physp,
				osm_physp_get_port_num(p_physp),
				payload, FALSE,
				ib_port_info_get_m_key(&p_physp->port_info),
				&context, TRUE);
	if (status != IB_SUCCESS)
		ret = -1;
	/* If we sent a new mkey above, update our guid2mkey map
	   now, on the assumption that the SubnSet succeeds
	*/
	if (update_mkey)
		osm_db_guid2key_set(p_mgr->p_subn->p_g2m,
				    cl_ntoh64(p_physp->port_guid),
				    cl_ntoh64(p_pi->m_key));

Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return ret;
}

/**********************************************************************
 Processes our own node
 Lock must already be held.
**********************************************************************/
static int lid_mgr_process_our_sm_node(IN osm_lid_mgr_t * p_mgr)
{
	osm_port_t *p_port;
	uint16_t min_lid_ho;
	uint16_t max_lid_ho;
	int ret;

	OSM_LOG_ENTER(p_mgr->p_log);

	/*
	   Acquire our own port object.
	 */
	p_port = osm_get_port_by_guid(p_mgr->p_subn,
				      p_mgr->p_subn->sm_port_guid);
	if (!p_port) {
		OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0308: "
			"Can't acquire SM's port object, GUID 0x%016" PRIx64
			"\n", cl_ntoh64(p_mgr->p_subn->sm_port_guid));
		ret = -1;
		goto Exit;
	}

	/*
	   Determine the LID this SM will use for its own port.
	   Be careful.  With an LMC > 0, the bottom of the LID range becomes
	   unusable, since port hardware will mask off least significant bits,
	   leaving a LID of 0 (invalid).  Therefore, make sure that we always
	   configure the SM with a LID that has non-zero bits, even after
	   LMC masking by hardware.
	 */
	lid_mgr_get_port_lid(p_mgr, p_port, NULL,
			     p_mgr->p_subn->opt.master_sm_lid,
			     &min_lid_ho, &max_lid_ho);
	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
		"Current base LID is %u\n", min_lid_ho);
	/*
	   Update subnet object.
	 */
	p_mgr->p_subn->master_sm_base_lid = cl_hton16(min_lid_ho);
	p_mgr->p_subn->sm_base_lid = cl_hton16(min_lid_ho);

	OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE,
		"Assigning SM's port 0x%016" PRIx64
		"\n\t\t\t\tto LID range [%u,%u]\n",
		cl_ntoh64(osm_port_get_guid(p_port)), min_lid_ho, max_lid_ho);

	/*
	   Set the PortInfo the Physical Port associated with this Port.
	 */
	ret = lid_mgr_set_physp_pi(p_mgr, p_port, p_port->p_physp,
				   cl_hton16(min_lid_ho));


Exit:
	OSM_LOG_EXIT(p_mgr->p_log);
	return ret;
}

int osm_lid_mgr_process_sm(IN osm_lid_mgr_t * p_mgr)
{
	int ret;

	OSM_LOG_ENTER(p_mgr->p_log);

	OSM_ASSERT(p_mgr->p_subn->sm_port_guid);

	CL_PLOCK_EXCL_ACQUIRE(p_mgr->p_lock);

	/* initialize the port_lid_tbl and empty ranges list following the
	   persistent db */
	lid_mgr_init_sweep(p_mgr);

	ret = lid_mgr_process_our_sm_node(p_mgr);

	CL_PLOCK_RELEASE(p_mgr->p_lock);

	OSM_LOG_EXIT(p_mgr->p_log);
	return ret;
}

/**********************************************************************
 1 go through all ports in the subnet.
 1.1 call lid_mgr_get_port_min_lid
 1.2 if a change is required send the port info
 2 if any change send the signal PENDING...
**********************************************************************/
int osm_lid_mgr_process_subnet(IN osm_lid_mgr_t * p_mgr)
{
	cl_qmap_t *p_port_guid_tbl;
	osm_port_t *p_port;
	ib_net64_t port_guid;
	int lid_changed, ret = 0;
	uint16_t min_lid_ho, max_lid_ho, max_flid;
	uint16_t lids, base_lid;

	OSM_ASSERT(p_mgr);

	OSM_LOG_ENTER(p_mgr->p_log);

	CL_PLOCK_EXCL_ACQUIRE(p_mgr->p_lock);

	OSM_ASSERT(p_mgr->p_subn->sm_port_guid);

	p_port_guid_tbl = &p_mgr->p_subn->port_guid_tbl;

	for (p_port = (osm_port_t *) cl_qmap_head(p_port_guid_tbl);
	     p_port != (osm_port_t *) cl_qmap_end(p_port_guid_tbl);
	     p_port = (osm_port_t *) cl_qmap_next(&p_port->map_item)) {
		port_guid = osm_port_get_guid(p_port);

		/*
		   Our own port is a special case in that we want to
		   assign a LID to ourselves first, since we have to
		   advertise that LID value to the other ports.

		   For that reason, our node is treated separately and
		   we will not add it to any of these lists.
		 */
		if (port_guid == p_mgr->p_subn->sm_port_guid) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG,
				"Skipping our own port 0x%016" PRIx64 "\n",
				cl_ntoh64(port_guid));
			continue;
		}

		/*
		   get the port lid range - we need to send it on first active
		   sweep or if there was a change (the result of
		   lid_mgr_get_port_lid)
		 */
		lid_changed = lid_mgr_get_port_lid(p_mgr, p_port, NULL, 0,
						   &min_lid_ho, &max_lid_ho);

		/* we can call the function to update the port info as it known
		   to look for any field change and will only send an updated
		   if required */
		OSM_LOG(p_mgr->p_log, OSM_LOG_VERBOSE,
			"Assigned port 0x%016" PRIx64 ", %s LID [%u,%u]\n",
			cl_ntoh64(port_guid), lid_changed ? "new" : "",
			min_lid_ho, max_lid_ho);

		/*
		 * If we change LID of an old port, then we need to invalidate
		 * ucast cache.
		 */
		if (p_mgr->p_subn->first_time_master_sweep == FALSE && lid_changed &&
		    !p_port->is_new) {
			base_lid = cl_ntoh16(osm_port_get_base_lid(p_port));

			/* Report switch rebooted */
			if ((osm_node_get_type(p_port->p_node) == IB_NODE_TYPE_SWITCH) &&
			    (base_lid == 0 || base_lid == 0xffff)) {
				osm_report_notice_1312(p_mgr->p_subn, port_guid,
						       osm_node_get_node_guid(p_port->p_node),
						       osm_physp_get_port_num(p_port->p_physp));
				OSM_LOG(p_mgr->p_log, OSM_LOG_INFO,
					"Node 0x%016" PRIx64 ", Port 0x%016"
					PRIx64 " was rebooted\n", cl_ntoh64(port_guid),
					cl_ntoh64(osm_node_get_node_guid(p_port->p_node)));
			} else {
				OSM_LOG(p_mgr->p_log,
					OSM_LOG_INFO,
					"Lid [%u,%u] changed for existing port 0x%016"
					 PRIx64 " \n",
					 min_lid_ho, max_lid_ho,
					 cl_ntoh64(port_guid));
			}

			if (p_mgr->p_subn->opt.use_ucast_cache) {
				p_mgr->sm->ucast_mgr.variables_invalidate_cache = TRUE;
			}
		}

		/* the proc returns the fact it sent a set port info */
		if (lid_mgr_set_physp_pi(p_mgr, p_port, p_port->p_physp,
					 cl_hton16(min_lid_ho)))
			ret = -1;
	}			/* all ports */

	/* store the guid to lid table in persistent db */
	osm_db_store(p_mgr->p_g2l, p_mgr->p_subn->opt.fsync_high_avail_files);

	lids = cl_ptr_vector_get_size(&p_mgr->p_subn->port_lid_tbl);
	lids = lids ? lids - 1 : 0;
	max_flid = p_mgr->max_flid_assigned;
	lids = (max_flid > lids) ? max_flid : lids;

	if (p_mgr->p_subn->max_lid < lids)
		p_mgr->p_subn->max_lid = lids;

	CL_PLOCK_RELEASE(p_mgr->p_lock);

	OSM_LOG_EXIT(p_mgr->p_log);
	return ret;
}

ib_net16_t osm_lid_mgr_get_vport_lid(IN osm_lid_mgr_t * p_mgr,
				     IN osm_port_t * p_port,
				     IN osm_vport_t * p_vport)
{
	uint16_t min_lid, max_lid;
	ib_net16_t base_lid = osm_port_get_base_lid(p_port);

	if (p_vport->index) {
		lid_mgr_get_port_lid(p_mgr, p_port, p_vport, 0, &min_lid, &max_lid);
		if (min_lid) {
			OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Assigned lid %u "
				"for VPort 0x%" PRIx64 " (VPort %u of port 0x%"
				PRIx64 ")\n",
				min_lid, cl_ntoh64(osm_vport_get_guid(p_vport)),
				p_vport->index,
				cl_ntoh64(osm_port_get_guid(p_port)));
			osm_vlid_mgr_set_vlid(&p_mgr->vlid_mgr, min_lid,
					      cl_ntoh16(base_lid));
			return cl_hton16(min_lid);
		} else {
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0318: "
				"Failed to assign lid for for VPort 0x%"
				PRIx64 " (VPort %u of port 0x%" PRIx64 ")\n",
				cl_ntoh64(osm_vport_get_guid(p_vport)),
				p_vport->index,
				cl_ntoh64(osm_port_get_guid(p_port)));
			return 0;
		}
	} else
		return base_lid;
}

void osm_lid_mgr_clear_vport_lid(IN osm_lid_mgr_t * p_mgr,
				 IN osm_port_t * p_port,
				 IN osm_vport_t * p_vport,
				 IN boolean_t clear_guid2lid)
{
	uint64_t guid = cl_ntoh64(osm_vport_get_guid(p_vport));
	uint16_t min_lid, max_lid, lid;
	boolean_t guid2lid;

	if (osm_db_guid2lid_get(p_mgr->p_g2l, guid, &min_lid, &max_lid)) {
		/* Lid not found in guid2lid db*/
		guid2lid = FALSE;
		lid = cl_ntoh16(osm_vport_get_lid(p_vport));

		/*
		 * If clear_guid2lid, function is called without checking if
		 * already exists, in order to avoid double search. Hence missing
		 * LID is not an error.
		 */
		if (lid && !clear_guid2lid)
			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 0319: "
				"VPort 0x%" PRIx64 " (VPort %u of port 0x%"
				PRIx64 ") dont have a lid in guid2lid db, "
				"using lid %u from vport info\n",
				cl_ntoh64(osm_vport_get_guid(p_vport)),
				p_vport->index,
				cl_ntoh64(osm_port_get_guid(p_port)), lid);
		else {
			if (clear_guid2lid)
				return;

			OSM_LOG(p_mgr->p_log, OSM_LOG_ERROR, "ERR 031A: "
				"No lid for VPort 0x%" PRIx64 " (VPort %u of "
				"port 0x%" PRIx64 ") in guid2lid db and vport "
				"info, Nothing to clear\n",
				cl_ntoh64(osm_vport_get_guid(p_vport)),
				p_vport->index,
				cl_ntoh64(osm_port_get_guid(p_port)));
			return;
		}
	} else {
		/* Lid found in guid2lid db*/
		guid2lid = TRUE;
		lid = min_lid;
	}

	OSM_LOG(p_mgr->p_log, OSM_LOG_DEBUG, "Clearing lid %u "
		"for VPort 0x%" PRIx64 " (VPort %u of port 0x%"
		PRIx64 "), Clear Guid2Lid: %s\n",
		lid, guid, p_vport->index,
		cl_ntoh64(osm_port_get_guid(p_port)),
		clear_guid2lid ? "Yes" : "No");

	osm_vlid_mgr_clear_vlid(&p_mgr->vlid_mgr, lid);

	if (guid2lid && clear_guid2lid) {
		osm_db_guid2lid_delete(p_mgr->p_g2l, guid);
		p_mgr->used_lids[lid] = 0;
	}
}

boolean_t osm_lid_mgr_is_flid(IN osm_lid_mgr_t * p_mgr, IN uint16_t lid)
{
	/* check if this LID falls into the FLID global range */
	return (p_mgr->p_subn->opt.global_flid_start <= lid &&
		lid <= p_mgr->p_subn->opt.global_flid_end);
}
